from random import sample
from string import ascii_uppercase

from django.db import models
from django.urls import reverse

from captcha.image import ImageCaptcha


class Captcha(models.Model):
    LEN = 5
    secret = models.CharField(max_length=LEN)
    image = models.ImageField(blank=True, upload_to="syngo/")

    def get_absolute_url(self):
        return reverse("syngo:check", kwargs={"pk": self.pk})

    def create_image(self):
        name = f"{self.id}.png"
        data = ImageCaptcha().generate(self.secret)
        self.image.save(name, data)

    @staticmethod
    def generate():
        c = Captcha.objects.create(secret="".join(sample(ascii_uppercase, Captcha.LEN)))
        c.create_image()
        return c



from django.contrib.auth import get_user_model
from django.test import TestCase
from django.urls import reverse

import syngo
from syngo.models import Captcha


class SyngoTest(TestCase):
    def test_syngo(self):
        User = get_user_model()
        user = User.objects.create(username="toto", password="toto")

        # Test register
        self.assertEqual(len(syngo.list_accounts()), 1)
        self.assertEqual(syngo.register(user).status_code, 201)
        self.assertEqual(syngo.register(user).status_code, 200)
        self.assertEqual(syngo.register(user).status_code, 200)
        self.assertEqual(len(syngo.list_accounts()), 2)

        # Test shadow_ban
        self.assertEqual(syngo.list_accounts()[1]["shadow_banned"], 0)
        syngo.shadow_ban(user)
        self.assertEqual(syngo.list_accounts()[1]["shadow_banned"], 1)
        syngo.shadow_ban(user, unban=True)
        self.assertEqual(syngo.list_accounts()[1]["shadow_banned"], 0)

        # Test deactivate
        syngo.deactivate(user)
        self.assertEqual(len(syngo.list_accounts()), 1)

        # Test registration token
        ret = syngo.registration_token().json()
        self.assertEqual(len(ret["token"]), 16)
        self.assertEqual(ret["uses_allowed"], 1)
        self.assertEqual(ret["completed"], 0)

        # Test views
        r = self.client.get(reverse("syngo:generate"))
        generated = Captcha.objects.last()
        self.assertEqual(generated.get_absolute_url(), r.url)

        get = self.client.get(r.url)
        self.assertIn(generated.image.url, get.content.decode())

        ko = self.client.post(r.url, {"guess": "toto"})
        self.assertIn(generated.image.url, ko.content.decode())

        ok = self.client.post(r.url, {"guess": generated.secret})
        self.assertIn("Success !", ok.content.decode())



from datetime import datetime, timedelta

from django.conf import settings

import httpx


def req(url, method="get", auth=True, **kwargs):
    if auth:
        kwargs["headers"] = {"Authorization": f"Bearer {settings.SYNGO_ACCESS_TOKEN}"}
    return getattr(httpx, method)(f"{settings.SYNGO_MATRIX_URL}{url}", **kwargs)


def django_to_matrix(user):
    """Get the matrix id associated to a django user."""
    return f"@{user.username}:{settings.SYNGO_MATRIX_DOMAIN}"


def register(django_user):
    """Register a Django user into a Matrix homeserver."""
    # https://matrix-org.github.io/synapse/latest/admin_api/user_admin_api.html
    # create-or-modify-account
    user_id = django_to_matrix(django_user)
    return req(
        f"/_synapse/admin/v2/users/{user_id}",
        method="put",
        json={"displayname": str(django_user)},
    )


def list_accounts(guests=False):
    """List accounts on a Matrix homeserver."""
    # https://matrix-org.github.io/synapse/latest/admin_api/user_admin_api.html
    # list-accounts
    accounts = []
    next_token = 0
    while True:
        ret = req(
            "/_synapse/admin/v2/users",
            params={"from": next_token, "limit": 10, "guests": guests},
        ).json()
        if "users" not in ret:  # pragma: no cover
            raise ValueError(f"Invalid response: {ret}")
        accounts += ret["users"]
        if "next_token" in ret:
            next_token = ret["next_token"]
        else:
            break
    return accounts


def shadow_ban(django_user, unban=False):
    """Shadow-(un)ban an user."""
    # https://matrix-org.github.io/synapse/latest/admin_api/user_admin_api.html
    # controlling-whether-a-user-is-shadow-banned
    user_id = django_to_matrix(django_user)
    return req(
        f"/_synapse/admin/v1/users/{user_id}/shadow_ban",
        method="delete" if unban else "post",
    )


def deactivate(django_user):
    """Deactivate an account."""
    # https://matrix-org.github.io/synapse/latest/admin_api/user_admin_api.html
    # deactivate-account
    user_id = django_to_matrix(django_user)
    return req(
        f"/_synapse/admin/v1/deactivate/{user_id}", method="post", json={"erase": True}
    )


def registration_token():
    """Create a registration token."""
    tomorrow = datetime.now() + timedelta(days=1)
    tomorrow = int(tomorrow.timestamp() * 1000)

    return req(
        "/_synapse/admin/v1/registration_tokens/new",
        method="post",
        json={"uses_allowed": 1, "expiry_time": tomorrow},
    )



# Generated by Django 3.2.15 on 2022-08-26 22:13

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Captcha",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("secret", models.CharField(max_length=5)),
                ("image", models.ImageField(blank=True, upload_to="syngo/")),
            ],
        ),
    ]



from django.shortcuts import render, redirect, get_object_or_404

from . import registration_token
from .models import Captcha


def generate(request):
    return redirect("syngo:check", pk=Captcha.generate().pk)


def check(request, pk):
    captcha = get_object_or_404(Captcha, pk=pk)
    if request.POST:
        if request.POST.get("guess", "").lower() == captcha.secret.lower():
            captcha.delete()
            token = registration_token().json()["token"]
            return render(request, "syngo/success.html", {"token": token})
    return render(request, "syngo/check.html", {"captcha": captcha})



from django.conf import settings
from django.conf.urls.static import static
from django.urls import path

from . import views

app_name = "syngo"
urlpatterns = [
    path("captcha/", views.generate, name="generate"),
    path("captcha/<int:pk>", views.check, name="check"),
]
if settings.DEBUG:  # pragma: no cover
    urlpatterns += static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['syngo', 'syngo.migrations']

package_data = \
{'': ['*'], 'syngo': ['templates/syngo/*']}

install_requires = \
['Django>=3.2,<4.0', 'captcha>=0.4,<0.5', 'httpx>=0.23.0,<0.24.0']

setup_kwargs = {
    'name': 'syngo',
    'version': '0.2.0',
    'description': 'Manage Synapse from Django',
    'long_description': '# SynGo\n\n[![Tests](https://github.com/nim65s/syngo/actions/workflows/test.yml/badge.svg)](https://github.com/nim65s/syngo/actions/workflows/test.yml)\n[![Lints](https://github.com/nim65s/syngo/actions/workflows/lint.yml/badge.svg)](https://github.com/nim65s/syngo/actions/workflows/lint.yml)\n[![Release](https://github.com/nim65s/syngo/actions/workflows/release.yml/badge.svg)](https://pypi.org/project/syngo/)\n[![pre-commit.ci status](https://results.pre-commit.ci/badge/github/nim65s/syngo/main.svg)](https://results.pre-commit.ci/latest/github/nim65s/syngo/main)\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![codecov](https://codecov.io/gh/nim65s/syngo/branch/main/graph/badge.svg?token=BLGISGCYKG)](https://codecov.io/gh/nim65s/syngo)\n[![Maintainability](https://api.codeclimate.com/v1/badges/a0783da8c0461fe95eaf/maintainability)](https://codeclimate.com/github/nim65s/syngo/maintainability)\n[![PyPI version](https://badge.fury.io/py/syngo.svg)](https://badge.fury.io/py/syngo)\n\nManage Synapse from Django\n\n## Install\n\n```\npython3 -m pip install syngo\n```\n\n## Unit tests\n\n```\ndocker-compose -f test.yml up --exit-code-from tests --force-recreate --build\n```\n',
    'author': 'Guilhem Saurel',
    'author_email': 'guilhem.saurel@laas.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nim65s/syngo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
